# -*- coding: utf-8 -*-
__all__ = []

__version__ = "0.1.0"
__uri__ = ''
__author__ = 'Shota Miyazaki and collaborators'
__email__ = 'miyazaki@ir.isas.jaxa.jp'
__license__ = 'MIT'
__description__ = 'Auto-differentiable Microlensing Modeling Framework with JAX.'